import org.jboss.as.controller.client.ModelControllerClient
import org.jboss.dmr.ModelNode

/**
 * Author: jmartisk
 * Date: 1/27/12
 * Time: 10:39 AM
 */

public class EapCommonTools {

    static SERVER_START_TIMEOUT_MILLISECONDS = 50000 /* time in miliseconds to wait for acknowledgement log in server.log stating that the server has started. Used both in domain and standalone mode. */



    public static void waitUntilDomainEapHasStarted(String jbossHomeDir, Process p) {
        Long timeBegin = System.currentTimeMillis()
        while (!(new File("$jbossHomeDir/domain/servers/server-two/log/server.log").exists() && new File("$jbossHomeDir/domain/servers/server-one/log/server.log").exists())) {
            if ((System.currentTimeMillis() - timeBegin) > SERVER_START_TIMEOUT_MILLISECONDS) {
                stopDomainEAP(jbossHomeDir, p, 4)
                throw new Exception("file /domain/servers/server-two/log/server.log didn't appear within the reasonable timeout, maybe you have changed the server's logging configuration?")
            }
        }
        while (true) {
            if (getLastLog("$jbossHomeDir/domain/servers/server-two/log/server.log").contains("started (with errors) in")) {
                throw new Exception("Server server-two started with errors => will not collect metrics")
            }
            if (getLastLog("$jbossHomeDir/domain/servers/server-one/log/server.log").contains("started (with errors) in")) {
                throw new Exception("Server server-one started with errors => will not collect metrics")
            }
            if (getLastLog("$jbossHomeDir/domain/servers/server-two/log/server.log").contains("started in") &&
                    getLastLog("$jbossHomeDir/domain/servers/server-one/log/server.log").contains("started in"))
                return
            if ((System.currentTimeMillis() - timeBegin) > SERVER_START_TIMEOUT_MILLISECONDS) {
                throw new Exception("Server failed to start within reasonable timeout")
            }
        }
    }

    public static void waitUntilStandaloneEapHasStarted(String jbossHomeDir, Process p) {
        Long timeBegin = System.currentTimeMillis()
        while (!(new File("$jbossHomeDir/standalone/log/server.log").exists())) {
            if ((System.currentTimeMillis() - timeBegin) > SERVER_START_TIMEOUT_MILLISECONDS) {
                throw new Exception("file /standalone/log/server.log didn't appear within the reasonable timeout, maybe you have changed the server's logging configuration?")
            }
        }
        while (true) {
            if (getLastLog("$jbossHomeDir/standalone/log/server.log").contains("started (with errors) in")) {
                throw new Exception("Server started with errors => will not collect metrics")
            }
            if (getLastLog("$jbossHomeDir/standalone/log/server.log").contains("started in"))
                return
            if ((System.currentTimeMillis() - timeBegin) > SERVER_START_TIMEOUT_MILLISECONDS) {
                throw new Exception("Server failed to start within reasonable timeout")
            }
        }
    }

    public static String getLastLog(String filepath) {
        if (new FileReader(filepath).readLines().isEmpty())
            return ""
        return new FileReader(filepath).readLines().last()
    }

    public static String getFirstLog(String filepath) {
        return new FileReader(filepath).readLine()
    }

    public static Process runStandaloneEAP(String jbossHome, String standaloneConfig) {
        return (jbossHome + "/bin/standalone.sh --server-config=$standaloneConfig").execute()
    }

    public static Process runDomainEAP(String jbossHome, String domainConfig, String hostConfig) {
        return (jbossHome + "/bin/domain.sh --domain-config=$domainConfig --host-config=$hostConfig").execute()
    }

    public static void stopStandaloneEAP(String jbossHome, Process currentProcess, int retries) {
        Thread.sleep(2000)
        ModelControllerClient mcc = ModelControllerClient.Factory.create(InetAddress.getByName("localhost"), 9999)
        ModelNode op = new ModelNode()
        op.get("operation").set("shutdown")
        println "operation: " + op.toJSONString(true)
        ModelNode ret
        try {
            ret = mcc.execute(op)
            println "result: " + ret.toJSONString(true)
            assert ret.get("outcome").asString().equals("success")
            Thread.sleep(15000)
        } catch (Exception e) {
            e.printStackTrace();
            if (retries > 0) {
                mcc.close()
                mcc = null
                println "************************************ Attempt to stop EAP failed, trying again $retries more times.."
                stopStandaloneEAP(jbossHome, currentProcess, retries - 1)
                Thread.sleep(15000)
            }
        }
        if (mcc != null) {
            mcc.close()
        }

/*
        Process p = "$jbossHome/bin/jboss-admin.sh -c --command=:shutdown".execute()
        p.waitFor()
        currentProcess.waitFor()
        println "************ stopping EAP>>> $p.text"
        Thread.sleep(1500)
*/
    }

    public static void stopDomainEAP(String jbossHome, Process currentProcess, int retries) {
        Thread.sleep(2000)
        ModelControllerClient mcc = ModelControllerClient.Factory.create(InetAddress.getByName("localhost"), 9999)
        ModelNode op = new ModelNode()
        op.get("operation").set("shutdown")
        op.get("address").set("host", "master")
        println "operation: " + op.toJSONString(true)
        ModelNode ret
        try {
            ret = mcc.execute(op)
            println "result: " + ret.toJSONString(true)
            assert ret.get("outcome").asString().equals("success")
        } catch (Exception e) {
            e.printStackTrace();
            if (retries > 0) {
                mcc.close()
                mcc = null
                println "************************************ Attempt to stop EAP failed, trying again $retries more times.."
                stopDomainEAP(jbossHome, currentProcess, retries - 1)
            }
        }
        if (mcc != null) {
            mcc.close()
        }
        /*
        Process p = "$jbossHome/bin/jboss-admin.sh --controller=localhost:9999 -c --command=/host=master/:shutdown".execute()
        p.waitFor()
        currentProcess.waitFor()
        println "************ stopping EAP>>> $p.text"
        Thread.sleep(1500)*/
    }

    /**
     * unzip a fresh jboss eap distribution from the specified zip file
     * will unzip into the current directory
     * then it will chmod +x the needed binaries
     * @param zipfile the zip file to be used
     */
    public static void prepareDistribution(String zipfile, String folderName) {
        def ant = new AntBuilder()
        Thread.sleep(5000)
        ant.delete(dir: folderName,
                deleteonexit: true)
        ant.unzip(src: zipfile,
                dest: ".",
                overwrite: false)
        ant.chmod(file: "$folderName/bin/standalone.sh", perm: "+x")
        ant.chmod(file: "$folderName/bin/jboss-admin.sh", perm: "+x")
        ant.chmod(file: "$folderName/bin/domain.sh", perm: "+x")
    }

    /************************************** FUNCTIONS **************************************/
    public static int readStandaloneStartupTime(String jboss_home) {
        String firstLog = getFirstLog("$jboss_home/standalone/log/boot.log")
        String lastLog = getLastLog("$jboss_home/standalone/log/server.log")
        println "First log of server bootstrap process: $firstLog"
        println "Last log of server bootstrap process: $lastLog"
        Date startDate = Date.parse("H:m:s,S", firstLog.split(" ")[0])
        Date endDate = Date.parse("H:m:s,S", lastLog.split(" ")[0])
        println endDate.time - startDate.time
        return endDate.time - startDate.time
    }

    public static int readDomainStartupTime(String jboss_home) {
        String firstLog = getFirstLog("$jboss_home/domain/log/host-controller/boot.log")
        String lastLog = getLastLog("$jboss_home/domain/servers/server-two/log/server.log")
        println "First log of server bootstrap process: $firstLog"
        println "Last log of server bootstrap process: $lastLog"
        Date startDate = Date.parse("H:m:s,S", firstLog.split(" ")[0])
        Date endDate = Date.parse("H:m:s,S", lastLog.split(" ")[0])
        println endDate.time - startDate.time
        return endDate.time - startDate.time
    }

    public static int readStandaloneMemoryFootprint(String jboss_home, int REPEATS_FOR_ONE_STARTUP, int DELAY_BETWEEN_MEASURING_SECONDS) {

        long total_total = 0
        REPEATS_FOR_ONE_STARTUP.times {
            int current =  getCurrentStandaloneMemoryFootprint()
            println "****Total memory used: ${current / (1024 * 1024)} MB"
            total_total +=  current
            Thread.sleep(DELAY_BETWEEN_MEASURING_SECONDS * 1000)
        }
        def average = Math.round((double) total_total / (REPEATS_FOR_ONE_STARTUP * 1024 * 1024))
        println "############### Average from $REPEATS_FOR_ONE_STARTUP measures: $average MB"

        return average
    }

    public static int getCurrentStandaloneMemoryFootprint() {
        ModelControllerClient mcc = ModelControllerClient.Factory.create(InetAddress.getByName("localhost"), 9999)
        ModelNode op = new ModelNode()
        op.get("address").add("core-service", "platform-mbean")
        op.get("address").add("type", "memory-pool")
        op.get("operation").set("read-resource")
        op.get("recursive").set("true")
        ModelNode ret = mcc.execute(op)
        int edenSpace = ret.get("result").get("name").get("PS_Eden_Space").get("usage").get("used").asInt()
        int oldGen = ret.get("result").get("name").get("PS_Old_Gen").get("usage").get("used").asInt()
        int permGen = ret.get("result").get("name").get("PS_Perm_Gen").get("usage").get("used").asInt()
        int survivorSpace = ret.get("result").get("name").get("PS_Survivor_Space").get("usage").get("used").asInt()
        int total = edenSpace + oldGen + permGen + survivorSpace
        mcc.close()
        return total
    }

    public static int readDomainMemoryFootprint(String jboss_home, int REPEATS_FOR_ONE_STARTUP, int DELAY_BETWEEN_MEASURING_SECONDS) {
        ModelControllerClient mcc = ModelControllerClient.Factory.create(InetAddress.getByName("localhost"), 9999)
        long total_total = 0
        REPEATS_FOR_ONE_STARTUP.times {
            ModelNode op = new ModelNode()
            op.get("address").add("host", "master")
            op.get("address").add("core-service", "platform-mbean")
            op.get("address").add("type", "memory-pool")
            op.get("operation").set("read-resource")
            op.get("recursive").set("true")
            ModelNode ret = mcc.execute(op)
            int edenSpace = ret.get("result").get("name").get("PS_Eden_Space").get("usage").get("used").asInt()
            int oldGen = ret.get("result").get("name").get("PS_Old_Gen").get("usage").get("used").asInt()
            int permGen = ret.get("result").get("name").get("PS_Perm_Gen").get("usage").get("used").asInt()
            int survivorSpace = ret.get("result").get("name").get("PS_Survivor_Space").get("usage").get("used").asInt()
            int total = edenSpace + oldGen + permGen + survivorSpace
            println "Eden space: $edenSpace"
            println "old generation: $oldGen"
            println "perm generation: $permGen"
            println "Survivor space: $survivorSpace"
            println "******************** Total memory used: ${total / (1024 * 1024)} MB"
            total_total += total
            Thread.sleep(DELAY_BETWEEN_MEASURING_SECONDS * 1000)
        }
        def average = Math.round((double) total_total / (REPEATS_FOR_ONE_STARTUP * 1024 * 1024))
        println "############### Average from $REPEATS_FOR_ONE_STARTUP measures: $average MB"
        mcc.close()
        return average
    }

    public static void deleteStandaloneLogFiles(String jbossHome) {
        try {
            new File("$jbossHome/standalone/logs/boot.log").delete();
            new File("$jbossHome/standalone/logs/server.log").delete();
        } catch (Exception e) {}
    }

    public static void deleteDomainLogFiles(String jbossHome) {
        try {
            new File("$jbossHome/domain/servers/server-two/log/server.log").delete();
            new File("$jbossHome/domain/servers/server-one/log/server.log").delete();
            new File("$jbossHome/domain/log/host-controller/boot.log").delete();
        } catch (Exception e) {
            println e.toString()
        }
    }


    public static int performDeploymentAndMeasureTime(String jbossHome, String deploymentFile) {
        println "deploying file: $deploymentFile"
        ModelControllerClient mcc = ModelControllerClient.Factory.create(InetAddress.getByName("localhost"), 9999)
        ModelNode op = new ModelNode()
        op.get("address").add("deployment", deploymentFile.substring(deploymentFile.lastIndexOf("/") + 1))
        op.get("operation").set("add")
        ModelNode list = new ModelNode()
        ModelNode x = new ModelNode()
        x.get("path").set(deploymentFile)
        x.get("archive").set(true)
        list.add(x)
        op.get("content").set(list)
        op.get("enabled").set(true)
        println "operation: " + op.toJSONString(true)
        ModelNode ret = mcc.execute(op)
        println "result: " + ret.toJSONString(true)
        assert ret.get("outcome").asString().equals("success")
        mcc.close()
        println "***************** SUCCESSFULLY deployed file $deploymentFile"
        return measureDeploymentTimeFor(jbossHome, deploymentFile.substring(deploymentFile.lastIndexOf("/") + 1))
    }

    public static String deployFile(String jbossHome, String deploymentFile, String runtimeName) {
        println "deploying file: $deploymentFile"
        ModelControllerClient mcc = ModelControllerClient.Factory.create(InetAddress.getByName("localhost"), 9999)
        ModelNode op = new ModelNode()
        String nametoReturn = deploymentFile.substring(deploymentFile.lastIndexOf("/") + 1)
        if (runtimeName != null) {
            op.get("runtime-name").set(runtimeName)
            nametoReturn = runtimeName
        }
        op.get("address").add("deployment", nametoReturn)
        op.get("operation").set("add")
        ModelNode list = new ModelNode()
        ModelNode x = new ModelNode()
        x.get("path").set(deploymentFile)
        x.get("archive").set(true)
        list.add(x)
        op.get("content").set(list)
        op.get("enabled").set(true)
        println "operation: " + op.toJSONString(true)
        ModelNode ret = mcc.execute(op)
        println "result: " + ret.toJSONString(true)
        assert ret.get("outcome").asString().equals("success")
        mcc.close()
        println "successfully deployed file $deploymentFile"
        return nametoReturn
    }

    public static String undeployFile(String jbossHome, String deploymentName) {
        println "undeploying file: $deploymentName"
        ModelControllerClient mcc = ModelControllerClient.Factory.create(InetAddress.getByName("localhost"), 9999)
        ModelNode op = new ModelNode()
        op.get("address").add("deployment", deploymentName)
        op.get("operation").set("remove")
        println "operation: " + op.toJSONString(true)
        ModelNode ret = mcc.execute(op)
        println "result: " + ret.toJSONString(true)
        assert ret.get("outcome").asString().equals("success")
        mcc.close()
        println "successfully undeployed file $deploymentName"
    }

    private static int measureDeploymentTimeFor(String jboss_home, String deploymentFileName) {
        println "finding logs for deployment filename: $deploymentFileName"
        String firstLog = getStartedDeploymentLogFor("$jboss_home/standalone/log/server.log", deploymentFileName)
        println "First log of deployment process: $firstLog"
        String lastLog = getFinishedDeploymentLogFor("$jboss_home/standalone/log/server.log", deploymentFileName)
        println "Last log of deployment process: $lastLog"
        Date startDate = Date.parse("H:m:s,S", firstLog.split(" ")[0])
        Date endDate = Date.parse("H:m:s,S", lastLog.split(" ")[0])
        println endDate.time - startDate.time
        return endDate.time - startDate.time
    }

    public static String getStartedDeploymentLogFor(String filepath, String filename) {
        while (true) {
//                println "searching for line containing:" + "Starting deployment of \"$filename\""
            String ret = null
            new FileReader(filepath).eachLine { line ->
                if (line.contains("Starting deployment of \"$filename\""))
                    ret = line
            }
            if (ret != null)
                return ret
        }
    }

    public static String getFinishedDeploymentLogFor(String filepath, String filename) {
        while (true) {
            String ret = null
            new FileReader(filepath).eachLine { line ->
                if (line.contains("Deployed \"$filename\"")) {
                    ret = line
                }
            }
            if (ret != null) {
                println "**************************** $ret"
                return ret
            }


        }
    }
}